#!/usr/bin/env python
"""
Create pipes and pipe networks using the Network class.

See the Readme for more details and examples.

Author: Duncan Hunter
"""

# pylint: disable=C0411
# pylint: disable=R0913
# pylint: disable=E1101

from pipemesh import pieces, gmsh
import os
import numpy as np
import xml.etree.ElementTree as ET
import collections

MODEL = gmsh.model
FACTORY = MODEL.occ
MESH = MODEL.mesh


def _round_0(values):
    """Rounds values less than 1e-8 to 0."""
    values = np.array(values)
    values[np.abs(values) < 1e-8] = 0
    return values


def _check_intersect(objects):
    """
    Check if entities in tools intersect entities in objects.

    Args:
        objects: (list) dimtags of volumes in model.

    Raises:
        ValueError: If pieces overlap.
    """
    for i in range(len(objects)):
        obj = objects[i]
        intersect = FACTORY.intersect([obj],
                                      objects[:i] + objects[i + 1:],
                                      removeObject=False,
                                      removeTool=False)[0]
        if intersect:
            raise ValueError("Pieces overlap")
    return


class Network():
    """Represents a pipe or network of pipes.

    Pipes are built from an inlet in a sequential, modular fashion.
    When a junction is added, a new "out surface" is added, which
    can be added to. In this way, a network of pipes can be built.

    Attributes:
        physical_in_out_surfaces: Dictionary of Physical surface tags
            to GMSH surface tags for inlets/outlets.
        phyiscal_no_slip: Dictionary of Physical surface tags for
            walls/outside of cylinder.
        physical_volume: Physical tag of the volume. Only available
            after generate.

    Methods:
        add_cylinder: Add a cylinder to the Network.
        add_curve: Add a curve to the Network.
        add_mitered: Add a mitered bend to the Network.
        add_change_radius: Add a cylinder with change in radius to the Network.
        add_t_junction: Add a T junction the the network.
        generate: Fuse the network and generates the msh file.

        get_inlet_outlet_phys_ids: Returns a list of physical ids of inlets.
        get_cyl_phys_ids: Returns a list of physical ids of cylinder surfaces.
        get_velocities_reynolds: Returns velocity vectors for inlets using
            Reynolds number.
        get_velocities_vel_mag: Returns velocity vectors for inlets using
            velocity magnitude.
    """

    def __init__(self, length, radius, direction, lcar=0.1):
        """
        Creates the inlet cylinder of a pipe.

        This is the beginning of a pipe, from which you can add more
        pieces. This piece has the inlet surface, and stores the
        geometrical information of the pipe, such as what direction
        the outlet is facing, where it is, and what the radius of the
        pipe is.

        Args:
        length: (float) length of piece.
        radius: (float) radius of pipe.
        direction: (list) Direction pipe will be facing in x, y, z
            vector format.
        lcar: (float) Mesh size of this piece. Maximum mesh size of model.
        """
        gmsh.initialize()
        if lcar <= 0:
            print("Defaulting to lcar of 0.1")
            lcar = 0.1
        gmsh.option.setNumber("Mesh.CharacteristicLengthMax", lcar)

        direction = np.array(direction)
        piece = pieces.Cylinder(length, radius, direction, lcar)
        piece.in_surface.direction *= -1

        self.piece_list = [piece]
        self.in_surfaces = [piece.in_surface]
        self.out_surfaces = [piece.out_surface]

        self.vol_tag = None  # Overall vol_tag, only after fuse.
        # dictionary of physical dim tags
        # to surface objects, which has information
        self.physical_in_out_surfaces = collections.OrderedDict()
        self.physical_no_slip = {}
        self.physical_volume = None

    def _set_mesh_sizes(self):
        """Sets the mesh size for all pieces."""
        field_list = []
        for piece in self.piece_list:
            half_length = np.abs(
                np.linalg.norm(piece.in_surface.centre - piece.vol_centre))
            field_length = np.linalg.norm(
                np.array([half_length, piece.in_surface.radius]))
            point = FACTORY.addPoint(*list(piece.vol_centre),
                                     meshSize=piece.lcar)
            FACTORY.synchronize()
            dist_field = MESH.field.add("Distance")
            MESH.field.setNumbers(dist_field, "NodesList", [point])
            thresh_field = MESH.field.add("Threshold")
            MESH.field.setNumber(thresh_field, "IField", dist_field)
            MESH.field.setNumber(thresh_field, "LcMin", piece.lcar)
            MESH.field.setNumber(thresh_field, "LcMax", 0.3)
            MESH.field.setNumber(thresh_field, "DistMin", field_length)
            MESH.field.setNumber(thresh_field, "DistMax", 1.1 * field_length)
            field_list.append(thresh_field)
        min_field = MESH.field.add("Min")
        MESH.field.setNumbers(min_field, "FieldsList", field_list)
        MESH.field.setAsBackgroundMesh(min_field)

    def _out_number(self, out_number):
        """Checks validity of out_number and changes to index form."""
        if out_number > len(self.out_surfaces):
            raise ValueError("Out piece does not exist")
        if out_number <= 0:
            out_number = 0
        else:  # Number to index
            out_number -= 1
        return out_number

    def add_cylinder(self, length, lcar=0.1, out_number=0):
        """Adds a pipe to the Network at the outlet.

        Args:
            length: (float) length of pipe.
            lcar: (float) mesh size of piece.
            out_number: Out surface to add to. If <= 1, will add to the
                first out surface.
        """
        if lcar <= 0:
            print("Defaulting to lcar of 0.1")
            lcar = 0.1
        out_number = self._out_number(out_number)
        out_surface = self.out_surfaces[out_number]
        piece = pieces.Cylinder(length, out_surface.radius,
                                out_surface.direction, lcar)
        translate_vector = out_surface.centre - piece.in_surface.centre
        FACTORY.translate([piece.vol_tag], *list(translate_vector))
        FACTORY.synchronize()
        piece._update_centres()
        self.piece_list.append(piece)
        self.out_surfaces[out_number] = piece.out_surface

    def add_curve(self, new_direction, bend_radius, lcar=0.1, out_number=0):
        """Adds a curve to the Network at the outlet.

        Args:
            new_direction: (list) Direction pipe will be facing
                in x, y, z vector format.
                e.g. [0, 1, 0] faces positive y.
            bend_radius: (float) Radius of the bend.
            lcar: (float) Size of mesh in this piece.
            out_number: Out surface to add to. If <= 1, will add to the
                first out surface.

        Raises:
            ValueError: new_direction vector isn't right size.
                Bend radius isn't big enough (<1.1 inlet radius).
        """
        # Check input
        if lcar <= 0:
            print("Defaulting to lcar of 0.1")
            lcar = 0.1
        out_number = self._out_number(out_number)
        out_surface = self.out_surfaces[out_number]
        if bend_radius < 1.1 * out_surface.radius:
            raise ValueError("""Bend radius is not large enough""")
        # Create Piece
        piece = pieces.Curve(out_surface.radius, out_surface.direction,
                             new_direction, bend_radius, lcar)
        # Translate Piece
        translate_vector = out_surface.centre - piece.in_surface.centre
        FACTORY.translate([piece.vol_tag], *list(translate_vector))
        FACTORY.synchronize()
        piece._update_centres()
        self.piece_list.append(piece)
        self.out_surfaces[out_number] = piece.out_surface

    def add_mitered(self, new_direction, lcar=0.1, out_number=0):
        """Adds a mitered bend to the Network at the outlet.

        A mitered bend is a sharp change in direction. Hard to
        simulate.

        Args:
            new_direction: (list, length 3) xyz vector representing
                the new direction of the pipe.
            lcar: (float) size of mesh of this piece.
            out_number: Out surface to add to. If <= 1, will add to the
                first out surface.
        """
        if lcar <= 0:
            print("Defaulting to lcar of 0.1")
            lcar = 0.1
        out_number = self._out_number(out_number)
        out_surface = self.out_surfaces[out_number]
        # Create Piece
        piece = pieces.Mitered(out_surface.radius, out_surface.direction,
                               new_direction, lcar)
        # Translate Piece
        translate_vector = out_surface.centre - piece.in_surface.centre
        FACTORY.translate([piece.vol_tag], *list(translate_vector))
        FACTORY.synchronize()
        piece._update_centres()
        self.piece_list.append(piece)
        self.out_surfaces[out_number] = piece.out_surface

    def add_change_radius(self,
                          length,
                          new_radius,
                          change_length,
                          lcar=0.1,
                          out_number=0):
        """Adds a piece that changes the radius of the outlet.

        The piece is length long, and changes the Network radius to
        new_radius, over change_length, which controls how gentle the
        change is.

        Args:
            length: (float) Length of the piece.
            new_radius: (float) radius to change to.
            change_length: (float) Length that the change takes
                place over. Must be less than length and > 0.
            lcar: (float) mesh size for this piece.
            out_number: Out surface to add to. If <= 1, will add to the
                first out surface.

        Raises:
            ValueErrors: change_length is not between length and 0.
                If radius does not change.
        """
        if lcar <= 0:
            print("Defaulting to lcar of 0.1")
            lcar = 0.1
        out_number = self._out_number(out_number)
        out_surface = self.out_surfaces[out_number]
        # Create Piece
        piece = pieces.ChangeRadius(length, change_length, out_surface.radius,
                                    new_radius, out_surface.direction, lcar)
        # Translate Piece
        translate_vector = out_surface.centre - piece.in_surface.centre
        FACTORY.translate([piece.vol_tag], *list(translate_vector))
        FACTORY.synchronize()
        piece._update_centres()
        self.piece_list.append(piece)
        self.out_surfaces[out_number] = piece.out_surface

    def add_t_junction(self, t_direction, lcar=0.1, t_radius=-1, out_number=0):
        """Adds a T junction to the Network at the outlet.

        This represents a pipe joining this pipe, creating a place to
        add a Network to this Network.

        Args:
            t_direction: (list, length 3) representing the direction
                that the joining pipe's inlet is facing.
            lcar: (float) mesh size for this piece.
            t_radius: radius of the piece joining the pipe. If <= 0, will
                default to radius of the pipe.
            out_number: Out surface to add to. If <= 1, will add to the
                first out surface.
        """
        if lcar <= 0:
            print("Defaulting to lcar of 0.1")
            lcar = 0.1
        out_number = self._out_number(out_number)
        out_surface = self.out_surfaces[out_number]

        if t_radius <= 0:
            t_radius = out_surface.radius
        # Create Piece
        piece = pieces.TJunction(out_surface.radius, t_radius,
                                 out_surface.direction, t_direction, lcar)
        # Translate Piece
        translate_vector = out_surface.centre - piece.in_surface.centre
        FACTORY.translate([piece.vol_tag], *list(translate_vector))
        FACTORY.synchronize()
        piece._update_centres()
        self.piece_list.append(piece)
        self.out_surfaces.append(piece.t_surface)
        self.out_surfaces[out_number] = piece.out_surface

    def _fuse_objects(self):
        """Fuses separate objects in Network together.

        Returns
            no_slip: (list) The dimtags of cylinder surfaces.
        """
        if len(self.piece_list) == 1:
            piece = self.piece_list[0]
            no_slip = [MODEL.getBoundary([piece.vol_tag], False)[0]]
            self.vol_tag = piece.vol_tag
            return no_slip

        if self.vol_tag:
            raise ValueError("Network already fused")

        vol_tags = [piece.vol_tag for piece in self.piece_list]

        # if _check_intersect([vol_tags[0]], vol_tags[1:]):
        _check_intersect(vol_tags)

        out_dim_tags = FACTORY.fuse([vol_tags[0]], vol_tags[1:])[0]
        FACTORY.synchronize()
        self.vol_tag = out_dim_tags[0]
        for piece in self.piece_list:
            piece.vol_tag = None
        surfaces = MODEL.getBoundary([self.vol_tag], False)
        tot_in = len(self.in_surfaces)
        tot_out = len(self.out_surfaces)
        found_in = 0
        found_out = 0
        no_slip = []
        for surf in surfaces:
            added = False
            loc = np.array(FACTORY.getCenterOfMass(*surf))
            if found_in < tot_in:
                for in_surf in self.in_surfaces:
                    if np.allclose(loc, in_surf.centre):
                        in_surf.dimtag = surf
                        found_in += 1
                        added = True
            if found_out < tot_out and not added:
                for out_surf in self.out_surfaces:
                    if np.allclose(loc, out_surf.centre):
                        out_surf.dimtag = surf
                        found_out += 1
                        added = True
            if not added:
                no_slip.append(surf)
        return no_slip

    def _set_physical_groups(self):
        """Sets the physical groups of the network.

        Sets every surface to a physical surface, and the volume
        to a physical volume."""
        no_slip = self._fuse_objects()

        for surface in self.in_surfaces + self.out_surfaces:
            phys_tag = MODEL.addPhysicalGroup(2, [surface.dimtag[1]])
            self.physical_in_out_surfaces[phys_tag] = surface

        for dimtag in no_slip:
            phys_tag = MODEL.addPhysicalGroup(2, [dimtag[1]])
            self.physical_no_slip[phys_tag] = dimtag
        self.physical_volume = MODEL.addPhysicalGroup(3, [self.vol_tag[1]])

    def get_inlet_outlet_phys_ids(self):
        """Returns a list of physical ids of inlets.

        By default, the inlet phys_id is 1,
        then the default outlet is 2, followed by any added outlets in the order
        they were added."""
        phys_ids = list(self.physical_in_out_surfaces.keys())
        return phys_ids

    def get_cyl_phys_ids(self):
        """Returns a list of physical ids of cylinder surfaces."""

        phys_ids = list(self.physical_no_slip.keys())
        return phys_ids

    def get_velocities_reynolds(self, physical_ids, reynolds_no, density,
                                viscosity):
        """Creates velocity vectors for inlets using reynolds number.

        Must be run after generate().
        Physical ids are turned into indices, which are used to select surfaces
        from self.physical_in_out_surfaces. By default, the inlet phys_id is 1,
        then the default outlet is 2, followed by any added outlets in the order
        they were added.
        Reynolds is turned into velocity mag using Reynolds*visc/(2*radius*density).

        Args:
            physical_ids: (list of ints) physical ids of inlets/outlet surfaces to
                get velocity vectors for.
            reynolds_no: (float) reynolds number resulting from velocity vectors
                created.
            density: (float) density of the fluid simulated. Should match density
                used in simulation for accurate results.
            viscosity: (float) viscosity of the fluid simulated. Should match
                viscosity used in simulation for accurate results.
        """
        if len(physical_ids) > len(self.physical_in_out_surfaces) - 1:
            raise ValueError(
                "Too many IDs given, at least one must be used for 0 pressure."
            )
        indices = np.array(physical_ids) - 1
        # Get information
        velocities = []
        for index in indices:
            surface = list(self.physical_in_out_surfaces.values())[index]
            radius = surface.radius
            direction = -surface.direction  # reversed to go into volume
            # find magnitude of direction
            mag = np.linalg.norm(direction)
            # find magnitude of velocity vector
            velo_mag = reynolds_no * viscosity / (2 * radius * density)
            # rescale direction to match velocity magnitude
            velocity = direction * (velo_mag / mag)
            # add to velocities
            velocities.append(velocity)
        return velocities

    def get_velocities_vel_mag(self, physical_ids, velocity_magnitude):
        """Returns velocity vectors for inlets using velocity magnitude.

        Must be run after generate().
        Physical ids are turned into indices, which are used to select surfaces
        from self.physical_in_out_surfaces. By default, the inlet phys_id is 1,
        then the default outlet is 2, followed by any added outlets in the order
        they were added."""
        if len(physical_ids) > len(self.physical_in_out_surfaces) - 1:
            raise ValueError(
                "Too many IDs given, at least one must be used for 0 pressure."
            )
        indices = np.array(physical_ids) - 1
        # Get information
        velocities = []
        for index in indices:
            surface = list(self.physical_in_out_surfaces.values())[index]
            direction = -surface.direction  # reversed to go into volume
            # find magnitude of direction
            mag = np.linalg.norm(direction)
            # rescale direction to match velocity magnitude
            velocity = direction * (abs(velocity_magnitude) / mag)
            # add to velocities
            velocities.append(velocity)
        return velocities

    def rotate_network(self, axis, angle):
        """Rotates the network from old_direction to new_direction.

        Args:
            axis: (array-like, shape (3,)) xyz vector representing the
                axis of rotation.
            angle: angle to rotate network about axis.
        """
        rot_axis = np.array(axis)
        dimtags = []
        for piece in self.piece_list:
            dimtags.append(piece.vol_tag)
        FACTORY.rotate(dimtags, 0, 0, 0, *list(rot_axis), angle)
        FACTORY.synchronize()
        for piece in self.piece_list:
            piece._update_centres()
            piece._update_directions(rot_axis, angle)

    def translate_network(self, vector):
        """Translates a network by vector.

        Args:
            vector: (list length 3) representing xyz vector to
                translate network by."""
        vector = np.array(vector)
        dimtags = []
        for piece in self.piece_list:
            dimtags.append(piece.vol_tag)
        FACTORY.translate(dimtags, *list(vector))
        FACTORY.synchronize()
        for piece in self.piece_list:
            piece._update_centres()

    def generate(self,
                 filename=None,
                 binary=False,
                 mesh_format='msh2',
                 write_info=False,
                 write_xml=False,
                 run_gui=False):
        """Generates mesh and saves if filename.

        Args:
            filename: (string) filename (without extension) to save as.
            binary: (Bool) Save mesh as binary or not. Default (False).
            mesh_format: (string) mesh format to save as. Default is
                msh2. To save as msh4, use 'msh4'.
            write_info: (Bool) write filename.txt with mesh
                mesh information (physical surfaces, locations
                and directions).
            write_xml: (Bool) write information in an xml file. Still
                under development.
            run_gui: (Bool) run the gmsh gui to view the mesh. May
                stop saving of information/meshes.
        """
        self._set_mesh_sizes()
        self._set_physical_groups()
        MESH.generate(3)
        # Set binary options
        if binary:
            gmsh.option.setNumber("Mesh.Binary", 1)
        else:
            gmsh.option.setNumber("Mesh.Binary", 0)
        # Set filename
        if filename:
            name = filename + "." + mesh_format
            gmsh.write(name)
            os.rename(name, filename + ".msh")
        if write_info:
            self._write_info(filename + ".txt")
        if write_xml:
            self._write_xml(filename)
        if run_gui:
            gmsh.option.setNumber("General.Axes", 2)
            gmsh.option.setNumber("Mesh.SurfaceFaces", 1)
            gmsh.fltk.run()
        gmsh.finalize()

    def _write_info(self, fname):
        """Writes network info into file fname.

        Writes Inlet/Outlet physical surface numbers, centres, and outwards
        normals. Then writes details about cylinders.
        """
        with open(fname, 'w') as myfile:
            myfile.writelines("Physical Surface, Centre, Outward Direction")
            myfile.writelines("\nInOut Surfaces")
            for key, surf in self.physical_in_out_surfaces.items():
                myfile.writelines("\n{}\t{}\t{}".format(
                    key, _round_0(surf.centre), _round_0(surf.direction)))
            myfile.writelines("\nCylinder Surfaces")
            for key, dimtag in self.physical_no_slip.items():
                centre = np.array(FACTORY.getCenterOfMass(2, dimtag[1]))
                myfile.writelines("\n{}\t{}".format(key, centre))
            myfile.writelines("\nIntersection locations and directions")
            for piece in self.piece_list:
                myfile.writelines("\n{}\t{}".format(
                    np.array(piece.out_surface.centre),
                    piece.out_surface.direction))
                if hasattr(piece, 't_surface'):
                    myfile.writelines("\n{}\t{}".format(
                        np.array(piece.t_surface.centre),
                        piece.t_surface.direction))
            myfile.close()

    def _write_xml(self, fname):
        """Writes information as xml tree."""
        root = ET.Element('root')
        inlet_surfs = ET.SubElement(root, "inlet_surfaces")
        for key, surf in self.physical_in_out_surfaces.items():
            surf = ET.SubElement(inlet_surfs,
                                 "{}".format(key),
                                 attrib={
                                     "centre":
                                     np.array2string(np.array(surf.centre)),
                                     "outward_direction":
                                     np.array2string(np.array(surf.direction))
                                 })
        cyl_surfs = ET.SubElement(root, "cylinder_surfaces")
        for key, dimtag in self.physical_no_slip.items():
            centre = np.array2string(
                np.array(FACTORY.getCenterOfMass(2, dimtag[1])))
            surf = ET.SubElement(cyl_surfs,
                                 "{}".format(key),
                                 attrib={"centre": centre})
        volume = ET.SubElement(root, "volume")
        volume.text = str(self.vol_tag)
        ET.ElementTree(root).write(fname + ".xml",
                                   encoding='utf-8',
                                   xml_declaration=True)



#!/usr/bin/env python
"""
Base classes for cylindrical GMSH pieces.

Also contains useful functions for these classes.
Author: Duncan Hunter
"""
# pylint: disable=C0411
# pylint: disable=R0913
from pipemesh import gmsh
import numpy as np
from scipy.spatial.transform import Rotation

MODEL = gmsh.model
FACTORY = MODEL.occ
MESH = MODEL.mesh


def vec_angle(vec1, vec2):
    """Returns the angle between two numpy array vectors"""
    return np.arccos(
        np.dot(vec1, vec2) / (np.linalg.norm(vec1) * np.linalg.norm(vec2)))


def proj(vec1, vec2):
    """
    Returns the component of vec1 along vec2

    Args:
        vec1, vec2: (np.array shape 3) xyz vector.
    """
    return np.dot(vec1, vec2) / np.linalg.norm(vec2)


def _rotate_inlet(vol_tag, in_direction, out_direction):
    """Rotates up facing inlet to face in_direction.

    Calculates the new outlet direction after it has been
    transformed.

    Args:
        vol_tag: dimtag tuple of volume being rotated.
        in_direction: xyz array direction to rotate the inlet to.
        out_direction: Direction the outlet is facing before
            rotation.

    Returns:
        new_out_direction: Direction outlet is facing after
            rotation as xyz array.
    """
    up_vector = np.array([0, 0, 1])
    # only have to rotate if its not facing up
    if np.allclose(in_direction, up_vector) is False:
        # Find rotation axis
        rotate_axis = np.cross(up_vector, in_direction)
        # If directions are parallel, set arbitrary rotate axis
        if np.allclose(rotate_axis, np.zeros(3)):
            rotate_axis = np.array([1, 0, 0])
        # Normalise axis for accurate scipy rotation
        rotate_axis = rotate_axis / np.linalg.norm(rotate_axis)
        rotate_angle = vec_angle(in_direction, up_vector)
        # GMSH rotation
        FACTORY.rotate([vol_tag], *[0, 0, 0], *list(rotate_axis), rotate_angle)
        FACTORY.synchronize()
        # Rotate out direction with scipy
        rot_vec = rotate_angle * rotate_axis
        rot1 = Rotation.from_rotvec(rot_vec)
        new_out_direction = rot1.apply(out_direction)
    else:
        new_out_direction = out_direction
    return new_out_direction


def _rotate_outlet(vol_tag, out_direction, in_direction, new_out_direction):
    """Rotates outlet about in_direction to face out_direction.

        Projects new_out_direction and out_direction onto basis axes
        that are perpendicular to each other and in_direction. The
        angle between the projections is found, and the object is
        rotated.

        Args:
            vol_tag: dimtag tuple of volume being rotated.
            out_direction: xyz array, the direction that the outlet
                will face after being rotated.
            in_direction: xyz array, the direction that the inlet is
                facing, and the axis that the object will be rotated
                about.
            new_out_direction: xyz array, the direction that the outlet
                faces before being rotated.
                Returned from _rotate_inlet.
        """
    basis_1 = np.cross(
        out_direction, in_direction
    )  # basis vectors perpendicular to direction (rotation axis)
    basis_2 = np.cross(basis_1,
                       in_direction)  # and perpendicular to other basis
    # Before rotation projection.
    alpha = np.array(
        [proj(new_out_direction, basis_1),
         proj(new_out_direction, basis_2)])
    # After rotation projection.
    beta = np.array(
        [proj(out_direction, basis_1),
         proj(out_direction, basis_2)])
    # Find angle between two vectors in bases.
    rot2_angle = vec_angle(alpha, beta)
    cross = np.cross(new_out_direction, out_direction)
    # Find direction to rotate in.
    if np.dot(in_direction, cross) > 0:
        rot2_angle *= -1
    # GMSH rotation.
    FACTORY.rotate([vol_tag], *[0, 0, 0], *list(in_direction), -rot2_angle)
    FACTORY.synchronize()


class Surface():
    """Class representing a surface of a piece.

    Pieces are PipePieces, and are all cylindrical in nature,
    which is why the pieces have radius.
    """

    def __init__(self, dimtag, centre, direction, radius):
        """Inits the surface."""
        self.dimtag = dimtag
        self.centre = centre
        self.direction = np.array(direction)
        self.radius = radius

    def _update_direction(self, axis, angle):
        """Rotates the direction vector by angle about axis."""
        rotvec = Rotation.from_rotvec(angle * axis)
        self.direction = rotvec.apply(self.direction)

    def _update_centre(self):
        self.centre = FACTORY.getCenterOfMass(*self.dimtag)


class PipePiece():
    """Parent class of pieces.

    Pieces are GMSH objects that can be used in creating pipes.
    This class has common information that all pieces have, such as
    radius. It also has functions that all the classes use, such as the
    need to update centres of pieces after they have been transformed.
    """

    def __init__(self, radius, vol_tag, in_tag, out_tag, in_direction,
                 out_direction, lcar):
        """Stores the information of a created piece.

        Args:
            radius: (float) radius of the piece.
            vol_tag: (tuple, length 2) GMSH dimtag (dimension, tag)
                representing volume.
            in_tag: (tuple, length 2) GMSH dimtag representing in surface
            out_tag: (tuple, length 2) GMSH dimtag representing out surface
            in_direction: (np array, shape 3) xyz vector representing
                direction going in.
            out_direction: (np array, shape 3) xyz vector representing
                direction going out.
            lcar: (float) mesh size of the piece.
        Raises:
            ValueError: Incorrect/Impossible dimensions.
        """
        if radius <= 0:
            raise ValueError("Radius must be greater than 0.")
        if np.linalg.norm(np.array(in_direction)) == 0:
            raise ValueError("Direction must be a vector with length.")
        if np.array(in_direction).shape[0] != 3:
            raise ValueError("3D vector needed.")
        self.lcar = lcar
        self.vol_tag = vol_tag
        self.vol_centre = np.array(FACTORY.getCenterOfMass(*vol_tag))
        in_centre = np.array(FACTORY.getCenterOfMass(*in_tag))
        out_centre = FACTORY.getCenterOfMass(*out_tag)
        self.in_surface = Surface(in_tag, in_centre, in_direction, radius)
        self.out_surface = Surface(out_tag, out_centre, out_direction, radius)

    def _update_surfaces(self):
        """Updates the dimtag of surfaces after a transformation.

        This function is overriden for every piece, as the order
        in which surfaces are assigned tags is different for different
        pieces."""
        surfaces = MODEL.getBoundary([self.vol_tag], combined=False)
        self.in_surface.dimtag = surfaces[2]
        self.out_surface.dimtag = surfaces[1]

    def _update_centres(self):
        """Updates centres of surfaces after a transformation."""
        self._update_surfaces()
        self.vol_centre = np.array(FACTORY.getCenterOfMass(*self.vol_tag))
        self.in_surface._update_centre()
        self.out_surface._update_centre()

    def _update_directions(self, axis, angle):
        """Updates the direction of surfaces after a transformation."""
        self.out_surface._update_direction(axis, angle)
        self.in_surface._update_direction(axis, angle)


class Cylinder(PipePiece):
    """
    Class representing a GMSH cylinder with base at 0,0,0 facing upwards.
    """

    def __init__(self, length, radius, direction, lcar):
        """Creates the cylinder with GMSH.

        Args:
            length: (float) length of cylinder.
            radius: (float) radius of cylinder.
            direction: (list, length 3) xyz vector representing direction
                cylinder is facing.
            lcar: (float) mesh size for this piece.
        Raises:
            ValueError: Incorrect/Impossible dimensions.
        """
        if length <= 0:
            raise ValueError("Length must be greater than 0.")
        self.length = length
        self.lcar = lcar
        # OpenCASCADE cylinder creation.
        vol_tag = (3, FACTORY.addCylinder(0, 0, 0, 0, 0, length, radius))
        FACTORY.synchronize()
        surfaces = MODEL.getBoundary([vol_tag], False)
        in_tag = surfaces[2]
        out_tag = surfaces[1]

        direction = np.array(direction)

        _rotate_inlet(vol_tag, direction, direction)

        super(Cylinder, self).__init__(radius, vol_tag, in_tag, out_tag,
                                       direction, direction, lcar)

    def _update_surfaces(self):
        """See base class."""
        surfaces = MODEL.getBoundary([self.vol_tag], combined=False)
        self.in_surface.dimtag = surfaces[2]
        self.out_surface.dimtag = surfaces[1]


class ChangeRadius(PipePiece):
    """Class representing a cylinder with a change in radius."""

    def __init__(self, length, change_length, in_radius, out_radius, direction,
                 lcar):
        """Inits the piece.

        Args:
            length: (float) length of piece.
            change_length: (float) length over which the change in
                radius takes place.
            in_radius: (float) radius of cylinder at "inlet".
            out_radius: (float) radius of cylinder at "outlet".
            direction: (list, length 3) xyz vector representing direction
                cylinder outlet is facing.
            lcar: (float) mesh size for this piece.
        Raises:
            ValueError: Incorrect/Impossible dimensions.
        """
        if change_length >= length:
            raise ValueError('change_length must be less than length.')
        if change_length <= 0:
            raise ValueError('change_length must be greater than 0.')
        if in_radius == out_radius:
            raise ValueError("Radius is not different.")
        if length <= 0:
            raise ValueError("length must be greater than 0.")
        # Add GMSH cylinder with largest radius.
        if out_radius > in_radius:
            vol_tag = (3, FACTORY.addCylinder(0, 0, 0, 0, 0, length,
                                              out_radius))
        else:
            vol_tag = (3, FACTORY.addCylinder(0, 0, 0, 0, 0, length,
                                              in_radius))
        FACTORY.synchronize()
        surfaces = MODEL.getBoundary([vol_tag], False)
        in_tag = surfaces[2]
        out_tag = surfaces[1]

        # Chamfer
        if out_radius > in_radius:
            lines = MODEL.getBoundary([in_tag], False, False)
            FACTORY.chamfer([vol_tag[1]], [lines[0][1]], [in_tag[1]],
                            [out_radius - in_radius, change_length])
            FACTORY.synchronize()
            self.increase = True
        else:
            lines = MODEL.getBoundary(out_tag, False, False)
            FACTORY.chamfer([vol_tag[1]], [lines[0][1]], [out_tag[1]],
                            [in_radius - out_radius, change_length])
            FACTORY.synchronize()
            self.increase = False

        direction = np.array(direction)

        _rotate_inlet(vol_tag, direction, direction)

        surfaces = MODEL.getBoundary([vol_tag], False)
        if self.increase:
            in_tag = surfaces[3]
            out_tag = surfaces[2]
        else:
            in_tag = surfaces[2]
            out_tag = surfaces[3]

        super(ChangeRadius, self).__init__(out_radius, vol_tag, in_tag,
                                           out_tag, direction, direction, lcar)

    def _update_surfaces(self):
        """See base class."""
        surfaces = MODEL.getBoundary([self.vol_tag], combined=False)
        if self.increase:
            self.in_surface.dimtag = surfaces[3]
            self.out_surface.dimtag = surfaces[2]
        else:
            self.in_surface.dimtag = surfaces[2]
            self.out_surface.dimtag = surfaces[3]


class Curve(PipePiece):
    """Class representing a GMSH curve by revolution."""

    def __init__(self, radius, in_direction, out_direction, bend_radius, lcar):
        """Inits the piece.

        Initially with inlet facing down, and outlet facing in x-plane.

        Args:
            radius: (float) radius of the pipe
            in_direction: (list, length 3) xyz vector representing
                direction going in.
            out_direction: (list, length 3) xyz vector representing
                direction going out.
            bend_radius: (float) radius of the bend of the curve.
            lcar: (float) mesh size for this piece.
        Raises:
            ValueError: Incorrect/Impossible dimensions.
        """
        in_tag = (2, FACTORY.addDisk(0, 0, 0, radius, radius))
        in_direction = np.array(in_direction)
        out_direction = np.array(out_direction)
        if np.allclose(in_direction, out_direction):
            raise ValueError("Directions must be different.")
        if np.linalg.norm(out_direction) == 0:
            raise ValueError("Out direction vector must have length.")
        if out_direction.shape[0] != 3:
            raise ValueError("3D vector needed.")
        revolve_axis = [0, 1, 0]
        centre_of_rotation = [bend_radius, 0, 0]
        angle = vec_angle(in_direction, out_direction)
        # Revolve in x plane, bend with radius bend_radius
        revolve_tags = FACTORY.revolve([in_tag], *centre_of_rotation,
                                       *revolve_axis, angle)
        FACTORY.synchronize()

        vol_tag = revolve_tags[1]

        # Direction outlet is facing
        new_out_direction = np.array(
            [np.sin(np.pi - angle), 0,
             -np.cos(np.pi - angle)])
        # Rotate so in_direction faces right way "Rot1"
        new_out_direction = _rotate_inlet(vol_tag, in_direction,
                                          new_out_direction)
        # Rotate so out_direction faces right way "Rot2"
        _rotate_outlet(vol_tag, out_direction, in_direction, new_out_direction)

        surfaces = MODEL.getBoundary([vol_tag], False, True)
        out_tag = surfaces[2]
        in_tag = surfaces[1]
        super(Curve, self).__init__(radius, vol_tag, in_tag, out_tag,
                                    in_direction, out_direction, lcar)

        # FACTORY.synchronize()
    def _update_surfaces(self):
        """See base class."""
        surfaces = MODEL.getBoundary([self.vol_tag], combined=False)
        self.in_surface.dimtag = surfaces[1]
        self.out_surface.dimtag = surfaces[2]


class Mitered(PipePiece):
    """Class representing a mitered (sharp) pipe bend.

    Piece creation is done by masking (intersect) a cylinder with a
    chamfered box. The piece is then mirrored, rotated and fused.

    The piece is then rotated to face the direction of the outflow.
    It is then rotated about the direction of outflow to match the new direction
    """

    def __init__(self, radius, in_direction, out_direction, lcar):
        """Creates the GMSH piece.

        The inlet is facing up originally.

        Args:
            radius: (float) radius of the pipe.
            in_direction: (list, length 3) xyz vector representing
            direction going in.
            out_direction: (list, length 3) xyz vector representing
            direction going out.
            lcar: (float) mesh size for this piece.
        Raises:
            ValueError: Incorrect/Impossible dimensions.
        """
        in_direction = np.array(in_direction)
        out_direction = np.array(out_direction)
        if np.allclose(in_direction, out_direction):
            raise ValueError("Directions must be different.")
        if np.linalg.norm(out_direction) == 0:
            raise ValueError("Out direction vector must have length.")
        if out_direction.shape[0] != 3:
            raise ValueError("3D vector needed.")
        # Chamfer cylinder
        angle = vec_angle(out_direction, in_direction)
        height = 2.1 * radius * np.tan(angle / 2)
        new_out_direction = np.array(
            [np.sin(np.pi - angle), 0,
             -np.cos(np.pi - angle)])  # original outlet direction in xz plane
        cyl1 = (3, FACTORY.addCylinder(0, 0, 0, 0, 0, height,
                                       radius))  # create cylinder
        box1 = (3,
                FACTORY.addBox(-radius - 1, -radius, -1, 2 * radius + 1,
                               2 * radius, height + 1))  # create box
        FACTORY.synchronize()
        surface = MODEL.getBoundary([box1], False,
                                    False)[5]  # Top surface to chamfer from
        line = MODEL.getBoundary([surface], False, False)[2]  # -x line on top
        sdist = 2 * radius * np.tan(
            angle / 2)  # distances to chamfer to get correct angle
        #ldist = 2*radius*np.cos(angle/2)
        FACTORY.chamfer([box1[1]], [line[1]], [surface[1]],
                        [2 * radius, sdist])  # chamfer
        int_tag = FACTORY.intersect([box1],
                                    [cyl1])  # intersect (chamfer on cylinder)
        fuse = FACTORY.fuse([int_tag[0]], int_tag[1:])[0]  # fuse
        fuse2 = FACTORY.copy([fuse])  # create copy and mirror
        FACTORY.symmetrize([fuse2], 1, 0, 0, 0)
        FACTORY.synchronize()
        surface = MODEL.getBoundary([fuse2], False,
                                    False)[1]  # get center of rotation
        com = FACTORY.getCenterOfMass(*surface)
        FACTORY.rotate([fuse2], *com, 0, 1, 0,
                       -(np.pi - angle))  # rotate to create piece
        vol_tag = FACTORY.fuse([fuse], [fuse2],
                               removeObject=True,
                               removeTool=True)[0][0]  # fuse
        FACTORY.synchronize()

        surfaces = MODEL.getBoundary([vol_tag], False)
        in_tag = surfaces[3]  # bottom
        out_tag = surfaces[6]  # angled

        # Rot1: rotate object so inlet faces correct direction
        new_out_direction = _rotate_inlet(vol_tag, in_direction,
                                          new_out_direction)
        #Rot2
        _rotate_outlet(vol_tag, out_direction, in_direction, new_out_direction)
        super(Mitered, self).__init__(radius, vol_tag, in_tag, out_tag,
                                      in_direction, out_direction, lcar)

    def _update_surfaces(self):
        """See base class."""
        surfaces = MODEL.getBoundary([self.vol_tag], combined=False)
        self.in_surface.dimtag = surfaces[3]
        self.out_surface.dimtag = surfaces[6]


class TJunction(PipePiece):
    """Class representing a T-junction in GMSH"""

    def __init__(self, radius, t_radius, direction, t_direction, lcar):
        """Inits the piece

        Creates a piece with the t_direction facing in the x-plane and
        inlet facing up.

        Args:
            radius: (float) radius of the pipe.
            t_radius: (float) radius of the joining pipe.
            direction: (list, length 3) xyz vector representing
                direction going in.
            t_direction: (list, length 3) xyz vector represeting the
                direction that the junction inlet faces.
            lcar: (float) mesh size for this piece.
        Raises:
            ValueError: Incorrect/Impossible dimensions.
        """
        if t_radius > radius:
            raise ValueError("t_radius cannot be bigger than radius")
        direction = np.array(direction)
        t_direction = np.array(t_direction)
        if np.allclose(direction, t_direction):
            raise ValueError("Directions must be different.")
        if np.linalg.norm(t_direction) == 0:
            raise ValueError("T direction vector must have length.")
        if t_direction.shape[0] != 3:
            raise ValueError("3D vector needed.")
        t_angle = vec_angle(direction, t_direction)
        # Create upside down or not. Ensures consistent surface assignment
        if t_angle > np.pi / 2:
            self.inv_surfs = True
            beta = abs(t_angle) - np.pi / 2
        else:
            self.inv_surfs = False
            beta = np.pi / 2 - abs(t_angle)

        height = radius * np.tan(beta) + radius / np.cos(beta)
        height_short = radius * abs(np.cos(beta))
        # Calculating height needed to emerge from merge

        # Add GMSH cylinders
        in_tag = (3, FACTORY.addCylinder(0, 0, 0, 0, 0, 1.1 * height, radius))
        mid_tag = (3, FACTORY.addCylinder(0, 0, 0, 1.1 * height, 0, 0,
                                          t_radius))
        out_tag = (3,
                   FACTORY.addCylinder(0, 0, 0, 0, 0, -1.1 * height_short,
                                       radius))

        # Rotate joining cylinder
        FACTORY.rotate([mid_tag], 0, 0, 0, 0, 1, 0, -beta)
        FACTORY.synchronize()
        # Fuse
        vol_tags = FACTORY.fuse([in_tag], [mid_tag, out_tag])[0]
        vol_tag = vol_tags[0]
        FACTORY.synchronize()

        surfaces = MODEL.getBoundary([vol_tag], False)
        # If created upside down (done to ensure order is the same)
        # Rotate the right way.
        if self.inv_surfs:
            FACTORY.rotate([vol_tag], 0, 0, 0, 1, 0, 0, np.pi)
            FACTORY.synchronize()
            mid_direction = np.array([np.cos(beta), 0, -np.sin(beta)])
            in_tag = surfaces[3]
            out_tag = surfaces[5]
        else:
            mid_direction = np.array([np.cos(beta), 0, np.sin(beta)])
            in_tag = surfaces[5]
            out_tag = surfaces[3]
        t_tag = surfaces[4]
        t_centre = FACTORY.getCenterOfMass(2, t_tag[1])

        mid_direction = _rotate_inlet(vol_tag, direction, mid_direction)

        _rotate_outlet(vol_tag, t_direction, direction, mid_direction)

        super(TJunction, self).__init__(radius, vol_tag, in_tag, out_tag,
                                        direction, direction, lcar)

        self.t_surface = Surface(t_tag, t_centre, t_direction, t_radius)

    def _update_surfaces(self):
        """See base class."""
        surfaces = MODEL.getBoundary([self.vol_tag], combined=False)
        if self.inv_surfs:
            self.in_surface.dimtag = surfaces[3]
            self.out_surface.dimtag = surfaces[5]
        else:
            self.in_surface.dimtag = surfaces[5]
            self.out_surface.dimtag = surfaces[3]
        self.t_surface.dimtag = surfaces[4]

    def _update_centres(self):
        """See base class."""
        self.vol_centre = np.array(FACTORY.getCenterOfMass(*self.vol_tag))
        self.in_surface.centre = np.array(
            FACTORY.getCenterOfMass(*self.in_surface.dimtag))
        self.out_surface.centre = np.array(
            FACTORY.getCenterOfMass(*self.out_surface.dimtag))
        self.t_surface.centre = np.array(
            FACTORY.getCenterOfMass(*self.t_surface.dimtag))

    def _update_directions(self, axis, angle):
        """See base class."""
        self.out_surface._update_direction(axis, angle)
        self.in_surface._update_direction(axis, angle)
        self.t_surface._update_direction(axis, angle)
